from esgcet.mapfile import ESGPubMapConv
from esgcet.mkd_non_nc import ESGPubMKDNonNC
from esgcet.update_solr import ESGUpdateSolr
from esgcet.update_globus import ESGUpdateGlobus

from esgcet.index_pub import ESGPubIndex
import sys
import esgcet.logger as logger

log = logger.ESGPubLogger()


class BasePublisher(object):

    def __init__(self, argdict):
        self.argdict = argdict
        self.fullmap = argdict["fullmap"]
        self.silent = argdict["silent"]
        self.verbose = argdict["verbose"]
        self.cert = argdict.get("cert","")
        self.index_node = argdict["index_node"]
        self.data_node = argdict["data_node"]
        self.data_roots = argdict["data_roots"]
        self.globus = argdict["globus"]
        self.replica = argdict["replica"]
        self.proj = argdict["proj"]
        self.json_file = argdict["json_file"]
        self.auth = argdict.get("auth",False)
        self.proj_config = argdict["user_project_config"]
        self.verify = argdict["verify"]
        self.mountpoints = argdict["mountpoints"]
        self.project = argdict["proj"]
        self.dry_run = argdict.get("dry_run", False)
        self.publog = log.return_logger('Generic Non-NetCDF Publisher', self.silent, self.verbose)

    def cleanup(self):
        pass

    def mapfile(self):

        mapconv = ESGPubMapConv(self.fullmap, project=self.project)
        map_json_data = None
        try:
            map_json_data = mapconv.mapfilerun(self.mountpoints)

        except Exception as ex:
            self.publog.exception("Failed to convert mapfile")
            self.cleanup()
            exit(1)
        return map_json_data

    def mk_dataset(self, map_json_data):
        mkd = ESGPubMKDNonNC(self.data_node, self.index_node, self.replica, self.globus, self.data_roots, 
                                self.silent, self.verbose)
        mkd.set_project(self.project)
        try:
            out_json_data = mkd.get_records(map_json_data, self.json_file, user_project=self.proj_config)
        except Exception as ex:
            self.publog.exception("Failed to make dataset")
            self.cleanup()
            exit(1)
        return out_json_data

    def update(self, json_data):
    
        if self.argdict.get("globus_index", False):
            up = ESGUpdateGlobus(self.argdict.get("index_UUID"), json_data[0]["data_node"], silent=self.silent, verbose=self.verbose, dry_run=self.dry_run)
        else:
            up = ESGUpdateSolr(self.index_node, silent=self.silent, verbose=self.verbose, verify=self.verify)
        try:
            up.run(json_data)
        except Exception as ex:
            self.publog.exception("Failed to update record")
            self.cleanup()
            exit(1)

    def index_pub(self,dataset_records):
        arch_cfg = None
        if self.argdict["enable_archive"]:
            arch_cfg = { "length" : int(self.argdict["archive_path_length"]) , 
                          "archive_path" : self.argdict["archive_path"]}
        print(f"VERBOSE: {self.verbose}")
        # TODO: support solr and Globus using the globus_index argument

        globuspub = self.argdict.get("globus_index", False)
        
        if globuspub:
            index_node = ""
        else:
            index_node = dataset_records[0]["index_node"]
        ip = ESGPubIndex(index_node=index_node, UUID=self.argdict["index_UUID"],  silent=self.silent, verbose=self.verbose, verify=self.verify, auth=self.auth, arch_cfg=arch_cfg, dry_run=self.dry_run)
            
        rc = True
        try:
            if globuspub:
                rc = ip.do_globus(dataset_records)
            else:
                rc = ip.do_publish(dataset_records)
        except Exception as ex:
            self.publog.exception("Failed to publish to index.")
            self.cleanup()
            exit(1)
        return rc

    def workflow(self):

        # step one: convert mapfile
        self.publog.info("Converting mapfile...")
        map_json_data = self.mapfile()

        # step two: make dataset
        self.publog.info("Making dataset...")
        out_json_data = self.mk_dataset(map_json_data)

        self.publog.info("Updating...")
        self.update(out_json_data)

        self.publog.info("Running index pub...")
        
        rc = self.index_pub(out_json_data)

        self.publog.info("Done.")

        return rc



from esgcet.handler_base import ESGPubHandlerBase
from datetime import datetime, timedelta
import json

class ESGPubAutocHandler(ESGPubHandlerBase):

    def get_scanfile_dict(self, scandata, mapdata):
#        self.publog.debug(json.dumps(scandata, indent=1))
        self.publog.debug(type(scandata['file']))
        ret = {}
        for key in scandata['file']:
            rec = scandata['file'][key]
            ret[rec['name']] = rec
        return ret

    def unpack_values(self, invals):
        """
        convert a dictionary of items under the key "values" to a list of the values

        invals (dict): input dictionary
        return list
        """
        for x in invals:
            if x['values']:
                yield x['values']

    def get_attrs_dict(self, scanobj):
        return scanobj['dataset']

    def get_variables(self, scanobj):
        return scanobj["variables"]
    
    def get_variable_list(self, variables):
        return list(variables.keys())

    def set_bounds(self, record, scanobj):
        geo_units = []
         
        if "axes" in scanobj:
            axes = scanobj["axes"]
            if "lat" in axes:
                lat = axes["lat"]
                geo_units.append(lat["units"])
                if 'values' not in lat.keys():
                    record["north_degrees"] = lat['subaxes']['0']["values"][-1]
                    record["south_degrees"] = lat['subaxes']['0']["values"][0]
                else:
                    record["north_degrees"] = lat["values"][-1]
                    record["south_degrees"] = lat["values"][0]
            if "lon" in axes:
                lon = axes["lon"]
                geo_units.append(lon["units"])
                if 'values' not in lon.keys():
                    record["east_degrees"] = lon['subaxes']['0']["values"][-1]
                    record["west_degrees"] = lon['subaxes']['0']["values"][0]
                else:
                    record["east_degrees"] = lon["values"][-1]
                    record["west_degrees"] = lon["values"][0]
            if "time" in axes:
                time_obj = axes["time"]
                time_units = time_obj["units"]
                tu_parts = []
                if type(time_units) is str:
                    tu_parts = time_units.split()
                if len(tu_parts) > 2 and tu_parts[0] == "days" and tu_parts[1] == "since":
                    proc_time = True
                    tu_date = tu_parts[2]  # can we ignore time component?
                    if "subaxes" in time_obj:
                        subaxes = time_obj["subaxes"]
                        sub_values = sorted([x for x in self.unpack_values(subaxes.values())])

                        tu_start_inc = int(sub_values[0][0])
                        tu_end_inc = int(sub_values[-1][-1])
                    elif "values" in time_obj:
                        tu_start_inc = time_obj["values"][0]
                        tu_end_inc = time_obj["values"][-1]
                    else:
                        self.publog.warning("Time values not located...")
                        proc_time = False
                    if proc_time:
                        try:
                            days_since_dt = datetime.strptime(tu_date.split("T")[0], "%Y-%m-%d")
                        except:
                            tu_date = '0' + tu_date
                            while len(tu_date.split('-')[0]) < 4:
                                tu_date = '0' + tu_date
                            days_since_dt = datetime.strptime(tu_date.split("T")[0], "%Y-%m-%d")
                        dt_start = days_since_dt + timedelta(days=tu_start_inc)
                        dt_end = days_since_dt + timedelta(days=tu_end_inc)
                        if dt_start.microsecond >= 500000:
                            dt_start = dt_start + timedelta(seconds=1)
                        dt_start = dt_start.replace(microsecond=0)
                        record["datetime_start"] = "{}Z".format(dt_start.isoformat())
                        record["datetime_end"] = "{}Z".format(dt_end.isoformat())

            if "plev" in axes:
                plev = axes["plev"]
                if "units" in plev and "values" in plev:
                    record["height_units"] = plev["units"]
                    record["height_top"] = plev["values"][0]
                    record["height_bottom"] = plev["values"][-1]
            if len(geo_units) > 0:
                record["geo_units"] = geo_units
        else:
            self.publog.warning("No axes extracted from data files")




import json
import requests
from datetime import datetime

from esgcet.pub_client import publisherClient
from esgcet.update_base import ESGUpdateBase
import esgcet.logger as logger

log = logger.ESGPubLogger()

class ESGUpdateSolr(ESGUpdateBase):

    def __init__(self, index_node : str, cert_fn : str = "", silent : bool = False, verbose : bool = False, verify : bool = False, auth : bool = False):
        self.index_node = index_node
        self.cert_fn = cert_fn
        self.silent = silent
        self.verbose = verbose
        self.verify = verify
        self.pubCli = publisherClient(self.cert_fn, self.index_node, verify=verify, verbose=self.verbose, silent=self.silent, auth=auth)

        self.SEARCH_TEMPLATE = 'https://{}/esg-search/search/?latest=true&distrib=false&format=application%2Fsolr%2Bjson&data_node={}&master_id={}&fields=version,id'
        self.publog = log.return_logger('Update Record', silent, verbose)


    def update_file(self, dsetid : str):
        self._update_core(dsetid, "files")


    def update_dataset(self, dsetid : str):
        self._update_core(dsetid,"datasets")


    def gen_hide_xml(self, id : str, core_type : str):
        ''' Generate the xml to hide the previous version

            id - the full dataset identifier
            core_type - which core to target: "files" or datasets"
        '''

        dateFormat = "%Y-%m-%dT%H:%M:%SZ"
        now = datetime.utcnow()
        ts = now.strftime(dateFormat)
        idfield = "id"
        if core_type == "files":
            idfield = "dataset_id"
        txt = f"""<updates core="{core_type}" action="set">
            <update>
              <query>{idfield}={id}</query>
              <field name="latest">
                 <value>false</value>
              </field>
              <field name="_timestamp">
                 <value>{ts}</value>
              </field>
            </update>
            </updates>
            \n"""

        return txt

    def query_update(self, dnode : str, mst : str):
        return self._query_esg_search(dnode, mst)

    def _query_esg_search(self, dnode : str, mst : str):

        # query for
        url = self.SEARCH_TEMPLATE.format(self.index_node, dnode, mst)

        self.publog.debug("Search Url: '{}'".format(url))
        resp = requests.get(url, verify=self.verify)

        self.publog.debug(resp.text)
        if not resp.status_code == 200:
            self.publog.error('Received {} from index server.'.format(resp.status_code))
            exit(1)

        res = json.loads(resp.text)

        if res['response']['numFound'] > 0:
            docs = res['response']["docs"]
            dsetid = docs[0]['id']
            return dsetid
        else:
            return ""

    def _update_core(self, dsetid : str, core_type : str):
        """ For a specific core, generate the xml and run the update.

            id - the full dataset identifier
            core_type - which core to target: "files" or datasets"
        """
        update_rec = self.gen_hide_xml(dsetid, core_type)
        self.publog.debug(update_rec)
        self.pubCli.update(update_rec)



from esgcet.mk_dataset_autoc import ESGPubAutocHandler
from esgcet.mk_dataset_xarray import ESGPubXArrayHandler
from esgcet.mk_dataset_nc4 import ESGPubNC4Handler

from esgcet.mk_dataset import ESGPubMakeDataset

import json, os, sys
import tempfile
from esgcet.generic_pub import BasePublisher
import traceback
import esgcet.logger as logger

log = logger.ESGPubLogger()

class GenericPublisher(BasePublisher):

    scan_file = tempfile.NamedTemporaryFile()  # create a temporary file which is deleted afterward for autocurator
    scanfn = scan_file.name

    def __init__(self, argdict):
        super().__init__(argdict)

        self.MKD_Construct = ESGPubMakeDataset
        if argdict["skipxr"]:
            self.autoc_command = None
            self.format_handler = ESGPubNC4Handler
            self.extract_method = self.nc4_load
        elif argdict["autoc_command"]:
            self.autoc_command = argdict["autoc_command"]
            self.format_handler = ESGPubAutocHandler
            self.extract_method = self.autocurator
        else:
            self.autoc_command = None
            self.extract_method = self.xarray_load
            self.format_handler = ESGPubXArrayHandler

        self.publog = log.return_logger('Generic NetCDF Publisher', self.silent, self.verbose)
        self._disable_further_info = argdict["disable_further_info"]

    def cleanup(self):
        self.scan_file.close()

    ## TODO: refactor these down to a single scan command
    def nc4_load(self, map_json_data):
        """
        netCDF4 LOAD
        """
        self.nc4_set = self.format_handler.nc4_load(map_json_data)

    def xarray_load(self, map_json_data):
        """
        Xarray Load
        """
        self.xarray_set = self.format_handler.xarray_load(map_json_data)


    def autocurator(self, map_json_data):
        """
        Autocurator
        """
        datafile = map_json_data[0][1]

        destpath = os.path.dirname(datafile)
        outname = os.path.basename(datafile)
        idx = outname.rfind('.')

        autstr = self.autoc_command + ' --out_pretty --out_json {} --files "{}/*.nc"'
        self.publog.debug(f"RUNNING {autstr}")
        stat = os.system(autstr.format(self.scanfn, destpath))
        if os.WEXITSTATUS(stat) != 0:
            self.publog.error("Autocurator exited with exit code: " + str(os.WEXITSTATUS(stat)))
            self.cleanup()
            exit(os.WEXITSTATUS(stat))

    def mk_dataset(self, map_json_data):
        
        https_url = self.argdict.get("https_url",None)    
        mkd = self.MKD_Construct(self.data_node, self.index_node, self.replica, self.globus, self.data_roots, 
                                 https_url, self.format_handler, self.silent, self.verbose, skip_opendap=self.argdict.get("skip_opendap",False))
        mkd.set_project(self.project)

        if self.argdict["skipxr"]:
            scan_arg = self.nc4_set
        elif self.autoc_command:
            scan_arg = json.load(open(self.scanfn))
        else:
            scan_arg = self.xarray_set

        try:
            out_json_data = mkd.get_records(map_json_data, scan_arg, self.json_file, user_project=self.proj_config)
        except Exception as ex:
            self.publog.exception("Failed to make dataset")
            self.cleanup()
            exit(1)
        return out_json_data

    def workflow(self):

        # step one: convert mapfile
        self.publog.info("Converting mapfile...")
        map_json_data = self.mapfile()

        # step two: autocurator
        self.publog.info(f"Running Extraction... {str(self.extract_method)}")
        self.extract_method(map_json_data)

        # step three: make dataset
        self.publog.info("Making dataset...")
        out_json_data = self.mk_dataset(map_json_data)

        # step four: update record if exists
        self.publog.info("Updating...")
        self.update(out_json_data)

        # step five: publish to database
        self.publog.info("Running index pub...")
        rc = self.index_pub(out_json_data)

        self.publog.info("Done. Cleaning up.")
        self.cleanup()
        return rc



from abc import ABC, abstractmethod

class ESGPubHandlerBase(ABC):
    def __init__(self, logger):
        self.publog = logger

    @abstractmethod
    def get_scanfile_dict(self, scandata, mapdata):
        """
        Get a dictionary associated with scanned data for the files, indexed by filename
        File-specific metadata, eg the tracking id of the file.  
        """
        pass

    @abstractmethod
    def get_attrs_dict(self, scanobj):
        """
        Return an a dictionary of the global attributes
        """
        pass

    @abstractmethod
    def get_variables(self, scanobj):
        """
        Returns a dictionary of variables, indexed by the names/ids of the variables
        """
        pass

    @abstractmethod
    def get_variable_list(self, variables):
        """
        returns the names/ids of the variables given a "varibales" object
        """
        pass

    @abstractmethod
    def set_bounds(self, record, scanobj):
        """
        Format-specific logic to set the bounds
        record (dict): the dataset record context 
        """
        pass